% function [x,Jit,FLAG] = sbr_ld(A,z,lambda);
% function [x,Jit,FLAG] = sbr_ld(A,z,lambda,VERB);
% function [x,Jit,FLAG] = sbr_ld(A,z,lambda,VERB,qinit);
% function [x,Jit,FLAG] = sbr_ld(A,z,lambda,VERB,qinit,explore);
% function [x,Jit,FLAG] = sbr_ld(A,z,lambda,VERB,qinit,explore,K);
% 
% Minimization of the cost function J(x) = ||z-Ax||_2^2 + lambda ||x||_0
% by the Single Best Replacement (SBR) algorithm
%
% This program is a supplementary material to the paper:
%
% From Bernoulli-Gaussian deconvolution to sparse signal restoration, 
% by Charles Soussen, Jerome Idier, David Brie, and Junbo Duan. 
% IEEE Trans. Signal Process., vol. 59, no. 10, pp. 4572-4584, Oct. 2011.
%
% Contact information: Charles.Soussen@univ-lorraine.fr
%----------------------------------------------------------------------
%
% Version which STORES THE GRAM MATRIX A'A and updates the Cholesky 
% factorization of (A'*A)(active_columns,active_columns). The 
% insertion/removal tests are done in parallel
%
%
% INPUTS
%
% - z: data signal, of size nz x 1
% - A: observation matrix of size nz x M (if A is sparse, then A'A is sparse)
% - lambda: sparsity parameter (scalar)
% - Optional argument VERB=1 for debug (many prints on screen), 0 by default
% - Optional argument qinit to choose a non-zero initialization
%   (qinit = [0;...;0] by default)
% - Optional argument explore: strategy of exploration of the culumns a_i
%
%   explore = 'all' (default): both insertions and removals are tested 
%   explore = 'remove': version which favors the removals: when a removal 
%                       decreases J(x), the "best removal" is selected
%                       and no insertion test is done. 
%   The 'remove' option may improve the approximation found with the
%   default option in some cases but the improvement does not occur 
%   systematically.
%
% - Optional argument K: SBR is stopped when an iterate has K non-zero
%   components. 
%   * When lambda > 0, this option shall not be activated since SBR 
%     terminates without any stopping condition.
%   * When lambda = 0, SBR coincides with Orthogonal Least Squares (OLS). 
%     Here, the K stopping condition is useful to stop OLS.
%
% 
% OUTPUTS:
%
%   - x (SBR output after full convergence): sparse vector of size Mx1, 
%     with M=size(A,2)
%   - Jit: vector of size (M+1)x1 or (K+1)x1, with M=size(A,2)
%     Save of the value of J(x) for each best iterate:
%     Jit(i) = the least J-value obtained with an iterate with i-1
%     entries (i>=1).
%     NOTE: in the version with qinit != [], Jit is not a full vector (there 
%     may be "holes" for early cardinalities since some values Jit(i) are 
%     lacking).
%   - FLAG: exit flag. 
%     Equal to 0 if the execution terminates correctly,
%     Equal to -1 if an anomaly occurs (numerical instability), i.e.
%     if J(x) < 0 or if there are numerical issues when computing the 
%     Cholesky factor of (A'*A)(active_columns,active_columns)
%
%---------------------------------------------------------------------

function [x,Jit,FLAG] = sbr_ld(A,z,lambda,varargin);

M = size(A,2);    % size of vector x

if (nargin == 3),
    VERB = 0;
    qinit = [];
    explore = 'all';
    K = M;
elseif (nargin == 4),
    VERB = varargin{1};
    qinit = [];
    explore = 'all';
    K = M;
elseif (nargin == 5),
    VERB = varargin{1};
    qinit = varargin{2};
    explore = 'all';
    K = M;
elseif (nargin == 6),
    VERB = varargin{1};
    qinit = varargin{2};
    explore = varargin{3};
    K = M;
elseif (nargin == 7),
    VERB = varargin{1};
    qinit = varargin{2};
    explore = varargin{3};
    K = varargin{4};
end;

K = min(K,min(size(A)));
    
% WE ASSUME THAT THE GRAM MATRIX A'A CAN BE COMPUTED AND STORED ONCE FOR ALL
%
AA = A'*A;
Az = A'*z;        % < h_i,z>
h2 = diag(AA);    % ||h_i||^2
z2 = z'*z;        % ||z||^2
Jit = spalloc(K+1,1,K+1);  % save of J(x) for the "best" iterates

% zero-valued initial solution
q = spalloc(M,1,M);    % sparse vector of size Mx1, initialized to 0
vecm1 = [];       % support = active indices 
lvecm1 = 0;       % size of support
lvecm0 = M;       % complement 
R = [];           % Cholesky factor of the Gram matrix:
                  % R is the upper triangular matrix such that 
                  %    AA(active_columns,active_columns) = R'*R 
                  % With the notations of [Soussen et al, 2011], R is the 
                  % transpose of L_Q (see Appendix C of [Soussen et al, 2011])
crit = z2;        % Jinit = J(0) = ||z||^2
FLAG = 1;         % Termination of SBR if FLAG = 0 (normal) or -1 (anomaly)
iter = 0;
Jit(1) = crit;
sol_cour = [];    % R'\Az(vecm1);

% USER INITIALIZATION 
if (~isempty(qinit)),
    q = qinit;
    vecm1 = find(qinit~=0);
    lvecm1 = length(vecm1);
    lvecm0 = M-lvecm1;
    if (lvecm1)
        R = full(chol(AA(vecm1,vecm1)));   % factorisation AA = R'R
        sol_cour = R'\Az(vecm1);
        crit = z2 - sol_cour'*sol_cour + lambda*lvecm1;
        Jit(lvecm1+1) = crit;
    end;
end;

if(VERB), fprintf('0:\t\t%1.3e\n',crit); end; 

% EXPLORATION STRATEGY: ALL COLUMNS OR NOT
if (strcmp(explore,'all')),  % all the columns are explored, starting by
                             % insertion tests
    while ((FLAG==1) & (lvecm1 < K)),
        % The current value of the cost function J(x) is already stored 
        % in crit
        dcritopt = 0;       % J_new - J_old
        mopt = [];          % position of the new column to insert
        indopt = [];        % location of the index to remove in vecm1
        mouv = 0;           % 1 if insertion, -1 if removal
        vecm0 = find(q==0); % updated at the beginning of the loop contrary
                            % to vecm1
        
        %--------------------------------------------------------------
        %-------- I. INSERTION/REMOVAL TESTS     ----------------------
        %--------------------------------------------------------------
        % INSERTION TRIALS h_m (m-th column) for all m
        if (lvecm1),        % non empty support
            tmp_vect = R'\AA(vecm1,vecm0);  % size lvecm1 x lvecm0
            F22_vect = h2(vecm0)-sum(tmp_vect.*tmp_vect,1)';
            dcritnew_vect = lambda - (tmp_vect'*sol_cour-Az(vecm0)).^2./F22_vect;
        else,   % empty support
            dcritnew_vect = lambda - Az.*Az./h2;
        end;            
        [val_min,ind] = min(dcritnew_vect);
        if (val_min < dcritopt),
            mouv = 1;
            mopt = vecm0(ind);   % the best insertion
            dcritopt = val_min;  % deltaJ(insert m) = the least deltaJ
        end;
        % END INSERTION TESTS
    
        % In some cases, it is not worth testing removals:
        if ((lambda > 0) & (dcritopt > -lambda) & (lvecm1>2)),
            % REMOVAL TESTS
            % Inversion of R matrix
            R_inv = inv(R);
            % Computation of current amplitudes 
            x_loc = R_inv*sol_cour;
            dcritnew_vect = x_loc.*x_loc./sum(R_inv.*R_inv,2) - lambda;
            [val_min,ind] = min(dcritnew_vect);

            if(val_min < dcritopt)
                mouv = -1;
                indopt = ind;       % index in VECM1 (not in q) of best removal
                dcritopt = val_min; 
            end;
            % END REMOVAL TESTS
        end;
        
        %--------------------------------------------------------------
        %-------- II. DO INSERTION, REMOVAL OR NOTHING    -------------
        %--------------------------------------------------------------
        if (dcritopt < 0),       % Modification of the support
            iter = iter + 1;
            crit = crit + dcritopt;

            if ((mouv == 1) & (lvecm1 > 0)),  % instability test
                tmp = R'\AA(vecm1,mopt);
                FLAG = 1-2*(h2(mopt)<tmp'*tmp);
            end;            

            if ((crit < -eps) | (FLAG == -1)),
                disp(sprintf('*** ABORT - NUMERICAL INSTABILITY, FLAG = %d ***\n',FLAG));
                FLAG = -1;
                % Amplitude computation
                x = spalloc(M,1,M); 
                x(vecm1) = R\sol_cour;
                return;
            end;
            
            if (mouv == 1),  % INSERTION
                % Update of R
                if (lvecm1 > 0),
                    R = [R, tmp];
                    R = [R; zeros(1,lvecm1), sqrt(h2(mopt)-tmp'*tmp)];
                else,
                    R = sqrt(h2(mopt));
                end;
                vecm1 = [vecm1; mopt]; % insertion of mopt at the last position
                lvecm0 = lvecm0 - 1;
                lvecm1 = lvecm1 + 1;
                %  note: vecm0 is updated in the beginning of the loop
                if(VERB), fprintf('%d:\t+%d\t%1.3e\n',iter,mopt,crit); end;
            else,          % REMOVAL
                % Update of R
                F = R(indopt+1:end,indopt+1:end);
                E = R(indopt,indopt+1:end);
                R(indopt,:) = [];
                R(:,indopt) = [];
                R(indopt:end,indopt:end) = cholupdate(full(F),full(E)');
                mopt = vecm1(indopt);
                vecm1(indopt) = [];  
                % vecm0 is updated in the beginning of the loop 
                lvecm0 = lvecm0 + 1;
                lvecm1 = lvecm1 - 1;
                if(VERB), fprintf('%d:\t-%d\t%1.3e\n',iter,mopt,crit); end;
            end;
            q(mopt) = ~q(mopt);   % 1-> 0 or 0 -> 1
            sol_cour = R'\Az(vecm1);
            Jit(lvecm1+1) = crit;
        else,
            FLAG = 0;   % Stop of SBR (no more decrease)
        end;
    end;  % WHILE
elseif (strcmp(explore,'remove')),  % the removals are first explored.
                                    % If one of them yields a decrease of
                                    % the cost, insertions are not tested
    %    
    while ((FLAG==1) & (lvecm1 < K)),
        % The current value of the cost function J(x) is already stored 
        % in crit
        dcritopt = 0;       % J_new - J_old
        mopt = [];          % position of the new column to insert 
        indopt = [];        % location of the index to remove in vecm1
        mouv = 0;           % 1 if insertion, -1 if removal
        vecm0 = find(q==0); % updated at the beginning of the loop contrary
                            % to vecm1
        
        %--------------------------------------------------------------
        %-------- I. INSERTION/REMOVAL TESTS     ----------------------
        %--------------------------------------------------------------
        if ((lambda > 0) & (lvecm1>2)),
            % REMOVAL TESTS 
            % Inversion of R matrix
            R_inv = inv(R);
            % Computation of current amplitudes 
            x_loc = R_inv*sol_cour;
            dcritnew_vect = x_loc.*x_loc./sum(R_inv.*R_inv,2) - lambda;
            [val_min,ind] = min(dcritnew_vect);
            if(val_min < dcritopt)
                mouv = -1;
                indopt = ind;  % index in VECM1 (not in q) of best removal
                dcritopt = val_min; 
            end;
            % END REMOVAL TESTS
        end;
        
        % If mouv=-1 the best removal is selected, else insertions are explored
        if (mouv == 0),
            % INSERTION TRIALS h_m (m-th column) for all m
            if (lvecm1),           % support of at least 2 elements
                tmp_vect = R'\AA(vecm1,vecm0);  % size lvecm1 x lvecm0
                F22_vect = h2(vecm0)-sum(tmp_vect.*tmp_vect,1)';
                dcritnew_vect = lambda - (tmp_vect'*sol_cour-Az(vecm0)).^2./F22_vect;
            else,   % empty support
                dcritnew_vect = lambda - Az.*Az./h2;
            end;            
            [val_min,ind] = min(dcritnew_vect);
            if(val_min < dcritopt)
                mouv = 1;
                mopt = vecm0(ind);  % the best insertion
                dcritopt = val_min; % deltaJ(insert m) = the least deltaJ
            end;
            % END INSERTION TESTS
        end;
        
        %--------------------------------------------------------------
        %-------- II. DO INSERTION, REMOVAL OR NOTHING    -------------
        %--------------------------------------------------------------
        if (dcritopt < 0),       % Modification of the support
            iter = iter + 1;
            crit = crit + dcritopt;
            if ((mouv == 1) & (lvecm1 > 0)),  % instability test 
                tmp = R'\AA(vecm1,mopt);
                FLAG = 1-2*(h2(mopt)<tmp'*tmp);
            end;            
            
            if ((crit < -eps) | (FLAG == -1)),
                disp(sprintf('*** ABORT - NUMERICAL INSTABILITY, FLAG = %d ***\n',FLAG));
                FLAG = -1;
                % Amplitude computation
                x = spalloc(M,1,M);
                x(vecm1) = R\sol_cour;
                return;
            end;

            if (mouv == 1),  % INSERTION
                % Update of R
                if (lvecm1 > 0),
                    R = [R, tmp];
                    R = [R; zeros(1,lvecm1), sqrt(h2(mopt)-tmp'*tmp)];
                else,
                    R = sqrt(h2(mopt));
                end;
                vecm1 = [vecm1; mopt]; % insertion of mopt at the last position
                lvecm0 = lvecm0 - 1;
                lvecm1 = lvecm1 + 1;
                %  note: vecm0 is updated in the beginning of the loop
                if(VERB), fprintf('%d:\t+%d\t%1.3e\n',iter,mopt,crit); end;
            else,          % REMOVAL
                % Update of R
                F = R(indopt+1:end,indopt+1:end);
                E = R(indopt,indopt+1:end);
                R(indopt,:) = [];
                R(:,indopt) = [];
                R(indopt:end,indopt:end) = cholupdate(full(F),full(E)');
                mopt = vecm1(indopt);
                vecm1(indopt) = [];
                % vecm0 is updated in the beginning of the loop 
                lvecm0 = lvecm0 + 1;
                lvecm1 = lvecm1 - 1;
                if(VERB), fprintf('%d:\t-%d\t%1.3e\n',iter,mopt,crit); end;
            end;
            q(mopt) = ~q(mopt);   % 1-> 0 or 0 -> 1
            sol_cour = R'\Az(vecm1);
            Jit(lvecm1+1) = crit;
        else,
            FLAG = 0;   % Stop of SBR (no more decrease)
        end;
    end; % WHILE
end; % IF

% Final amplitude computation
%
x = spalloc(M,1,M);
x(vecm1) = R\sol_cour;
